function _defineProperty(obj, key, value) {
  if (key in obj) {
    Object.defineProperty(obj, key, {
      value: value,
      enumerable: true,
      configurable: true,
      writable: true
    });
  } else {
    obj[key] = value;
  }

  return obj;
}

function _objectSpread(target) {
  for (var i = 1; i < arguments.length; i++) {
    var source = arguments[i] != null ? arguments[i] : {};
    var ownKeys = Object.keys(source);

    if (typeof Object.getOwnPropertySymbols === 'function') {
      ownKeys = ownKeys.concat(Object.getOwnPropertySymbols(source).filter(function (sym) {
        return Object.getOwnPropertyDescriptor(source, sym).enumerable;
      }));
    }

    ownKeys.forEach(function (key) {
      _defineProperty(target, key, source[key]);
    });
  }

  return target;
}

function _arrayWithHoles(arr) {
  if (Array.isArray(arr)) return arr;
}

function _iterableToArrayLimit(arr, i) {
  var _arr = [];
  var _n = true;
  var _d = false;
  var _e = undefined;

  try {
    for (var _i = arr[Symbol.iterator](), _s; !(_n = (_s = _i.next()).done); _n = true) {
      _arr.push(_s.value);

      if (i && _arr.length === i) break;
    }
  } catch (err) {
    _d = true;
    _e = err;
  } finally {
    try {
      if (!_n && _i["return"] != null) _i["return"]();
    } finally {
      if (_d) throw _e;
    }
  }

  return _arr;
}

function _nonIterableRest() {
  throw new TypeError("Invalid attempt to destructure non-iterable instance");
}

function _slicedToArray(arr, i) {
  return _arrayWithHoles(arr) || _iterableToArrayLimit(arr, i) || _nonIterableRest();
}

/*
 * Created Date: 2019-07-09
 * Author: 宋慧武
 * ------
 * Last Modified: Tuesday 2019-07-09 18:09:50 pm
 * Modified By: the developer formerly known as 宋慧武 at <songhuiwu001@ke.com>
 * ------
 * HISTORY:
 * ------
 * Javascript will save your soul!
 */
/**
 * @desc 获取 DOM CSS 属性的值
 *
 * @param {DOMElement} ele A DOM 元素
 * @returns {String}
 */

function getStylePropValue(ele, prop) {
  return window.getComputedStyle(ele).getPropertyValue(prop);
}
/**
 * @desc 元素是否隐藏
 *
 * @param {DOMElement} ele A DOM 元素
 * @returns {Boolean} true => 未隐藏可见  false => 隐藏不可见
 */

function isVisible(ele) {
  if (ele === window.document) {
    return true;
  }

  if (!ele || !ele.parentNode) {
    return false;
  }

  var parent = ele.parentNode;
  var visibility = getStylePropValue(ele, "visibility");
  var display = getStylePropValue(ele, "display");

  if (visibility === "hidden" || display === "none") {
    return false;
  }

  return parent ? isVisible(parent) : true;
}

/*
 * Created Date: 2019-07-11
 * Author: 宋慧武
 * ------
 * Last Modified: Sunday 2019-07-14 22:56:22 pm
 * Modified By: the developer formerly known as 宋慧武 at <songhuiwu001@ke.com>
 * ------
 * HISTORY:
 * ------
 * Javascript will save your soul!
 */

/**
 * @desc 获取对象的键值
 *
 * @param {Object} value
 * @returns {Array} [keys, values]
 */
function zipObject(value) {
  return [Object.values(value), Object.keys(value)];
}
/**
 * @desc 判断给定变量是否为一个函数
 *
 * @param {*} v
 * @return {Boolean}
 */

var isFun = function isFun(v) {
  return typeof v === "function" || false;
};
/**
 * @desc chenck 埋点ID是否合法
 *
 * @param {Object} value
 */

var vaildEvent = function vaildEvent(events, eventId) {
  if (!isFun(events[eventId])) {
    throw new Error("track eventId '".concat(eventId, "' does not exist"));
  }
};
/**
 * @desc check 异步埋点监听key是否存在
 *
 * @param {Object} value
 */

var vaildWatchKey = function vaildWatchKey(stateKey, propKey) {
  if (!stateKey && !propKey) {
    throw new Error("Missing arguments.{stateKey} or arguments.{propKey} in async track");
  }
};
/**
 * @desc check 当前实例是否为react组件
 */

var vaildReactComponent = function vaildReactComponent(isRC) {
  if (!isRC) {
    throw new Error("Missing arguments.{stateKey} or arguments.{propKey} in async track");
  } else {
    return isRC;
  }
};
/**
 * @desc 清除定时器
 */

var clearTimeoutQueue = function clearTimeoutQueue() {
  var queue = arguments.length > 0 && arguments[0] !== undefined ? arguments[0] : [];
  Object.keys(queue).forEach(function (timer) {
    clearTimeout(queue[timer]);
  });
};

var ONCE = "once";
var modifiers = {
  UVPV: "UVPV",
  // 特殊的修饰符
  TONP: "TONP",
  // 特殊的修饰符，表示页面停留时长
  CLICK: "click",
  CLICK_AFTER: "click.after",
  ASYNC: "async",
  ASYNC_DELAY: "async.delay"
};
/**
 * @desc inject 装饰器
 * @param {Function} reaction 如果为 Mobx Class 则需要注入 mobx.reaction 来监听通过异步action 对 state 的改变
 * @param {Object} trackEvents 当前页面需要的埋点事件
 */

var inject = function inject() {
  for (var _len = arguments.length, props = new Array(_len), _key = 0; _key < _len; _key++) {
    props[_key] = arguments[_key];
  }

  return function (target) {
    Object.assign.apply(Object, [target.prototype].concat(props));
  };
};
/**
 * @desc track 埋点装饰器
 * @param {String} modifier 修饰符，对应的埋点类型
 * @param {Number | String} eventId 埋点事件id
 * @param {Object} params 自定义参数，目前支持tateKey、propKey、delay、ref
 *
 * @property[stateKey] 对应组件 state 或者 mobx observable 中 key
 * @property[propKey] 对应组件 props 中 key
 * @property[delay] 埋点延迟时间
 * @property[ref] 对应 DOM 的引用，避免小于埋点延迟时间内发生DOM影藏造成埋点继续上报的问题
 */

function track(modifier, eventId) {
  var params = arguments.length > 2 && arguments[2] !== undefined ? arguments[2] : {};

  var _zipObject = zipObject(modifiers),
      _zipObject2 = _slicedToArray(_zipObject, 1),
      mdfs = _zipObject2[0];

  if (!mdfs.includes(modifier.replace(/\.once/g, ""))) {
    throw new Error("modifier '".concat(modifier, "' does not exist"));
  }

  return function (_, name, descriptor) {
    var handler;
    var value = descriptor.value,
        initializer = descriptor.initializer; // 页面初始化埋点

    if (modifier.includes(modifiers.UVPV)) {
      handler = function handler() {
        var _this = this;

        for (var _len2 = arguments.length, args = new Array(_len2), _key2 = 0; _key2 < _len2; _key2++) {
          args[_key2] = arguments[_key2];
        }

        vaildReactComponent(this.isReactComponent);
        var evts = this.trackEvents || this.props.trackEvents;
        var fn = value ? value.bind(this) : initializer.apply(this, args);

        var tck = function tck() {
          var _evts$modifier;

          var context = _objectSpread({}, _this.state, _this.props);

          (_evts$modifier = evts[modifier]).call.apply(_evts$modifier, [null, context].concat(args));
        };

        return [tck, fn].forEach(function (sub) {
          return sub.apply(void 0, args);
        });
      };
    } // 设置进入页面时间
    else if (modifier.includes(modifiers.TONP)) {
        handler = function handler() {
          var _this2 = this;

          var isRC = vaildReactComponent(this.isReactComponent); // 是否为 react 组件

          var evts = this.trackEvents || this.props.trackEvents;

          var tck = function tck() {
            var stt = Date.now() - _this2.__trackPageEntryTime;

            var context = _objectSpread({}, _this2.state, _this2.props);

            evts[modifier].call(null, context, {
              stt: stt
            });
          };

          this.__trackPageEntryTime = Date.now(); // 页面卸载前上报埋点

          if (isRC && !this.componentWillUnmount) {
            this.needMergeTONPWillUnmount = false;

            this.componentWillUnmount = function () {
              return tck();
            };
          } else if (this.componentWillUnmount && this.needMergeTONPWillUnmount !== false) {
            var willUnmountRef = this.componentWillUnmount;

            this.componentWillUnmount = function () {
              willUnmountRef.apply(_this2);
              tck();
            };
          }

          for (var _len3 = arguments.length, args = new Array(_len3), _key3 = 0; _key3 < _len3; _key3++) {
            args[_key3] = arguments[_key3];
          }

          return value ? value.apply(this, args) : initializer.apply(this, args).apply(void 0, args);
        };
      } // 事件行为埋点
      else if (modifier.includes(modifiers.CLICK)) {
          handler = function handler() {
            var _this3 = this;

            for (var _len4 = arguments.length, args = new Array(_len4), _key4 = 0; _key4 < _len4; _key4++) {
              args[_key4] = arguments[_key4];
            }

            var context = this;
            var isRC = this.isReactComponent; // 是否为 react 组件

            var evts = this.trackEvents || this.props.trackEvents;
            var once = modifier.includes(ONCE);
            var onceProp = "".concat(name, "_").concat(eventId);
            var after = modifier.includes(modifiers.CLICK_AFTER);

            var tck = function tck() {
              var _evts$eventId;

              if (_this3[onceProp]) return; // 如果存在once修饰符，且为true则直接返回

              vaildEvent(evts, eventId); // 检测eventId是否合法

              isRC && (context = _objectSpread({}, _this3.state, _this3.props));

              (_evts$eventId = evts[eventId]).call.apply(_evts$eventId, [null, context].concat(args));

              once && (_this3[onceProp] = true);
            };

            var fn = value ? value.bind(this) : initializer.apply(this, args);
            var queue = after ? [fn, isRC ? function () {
              return _this3.setState({}, tck);
            } : tck] : [tck, fn];
            return queue.forEach(function (sub) {
              return sub.apply(void 0, args);
            });
          };
        } // 异步行为埋点
        else {
            handler = function handler() {
              var _this4 = this;

              for (var _len5 = arguments.length, args = new Array(_len5), _key5 = 0; _key5 < _len5; _key5++) {
                args[_key5] = arguments[_key5];
              }

              var tck;
              var context = this;
              var isRC = this.isReactComponent; // 是否为 react 组件

              var evts = this.trackEvents || this.props.trackEvents;
              var once = modifier.includes(ONCE);
              var fn = value ? value.bind(this) : initializer.apply(this, args);
              var stateKey = params.stateKey,
                  propKey = params.propKey;
              var watchPropKey = "".concat(name, "_").concat(propKey, "_").concat(eventId); // 保证key的唯一性

              var watchStateKey = "".concat(name, "_").concat(stateKey, "_").concat(eventId);
              vaildEvent(evts, eventId);
              vaildWatchKey(stateKey, propKey);
              !this.tckQueue && (this.tckQueue = {}); // 在当前实例维护一个异步埋点队列

              !this.tckQueuePropKeys && (this.tckQueuePropKeys = []); //

              !this.tckQueueStateKeys && (this.tckQueueStateKeys = []); //

              if (modifier === modifiers.ASYNC_DELAY || modifier === modifiers.ASYNC_DELAY + ".once") {
                !this.tckTimerQueue && (this.tckTimerQueue = {}); // 定时器队列

                tck = function tck() {
                  var _params$delay = params.delay,
                      delay = _params$delay === void 0 ? 0 : _params$delay,
                      ref = params.ref;
                  var ele = _this4[ref] || document;
                  var timer = _this4.tckTimerQueue[eventId];
                  timer && clearTimeout(timer);
                  _this4.tckTimerQueue[eventId] = setTimeout(function () {
                    var _evts$eventId2;

                    isRC && (context = _objectSpread({}, _this4.state, _this4.props));
                    isVisible(ele) && (_evts$eventId2 = evts[eventId]).call.apply(_evts$eventId2, [null, context].concat(args));
                    clearTimeout(_this4.tckTimerQueue[eventId]);
                  }, delay);
                }; // 页面卸载时清除进行中的定时器


                if (isRC && !this.componentWillUnmount) {
                  this.needMergeWillUnmount = false;

                  this.componentWillUnmount = function () {
                    clearTimeoutQueue(_this4.tckTimerQueue);
                  };
                } else if (this.componentWillUnmount && this.needMergeWillUnmount !== false) {
                  var willUnmountRef = this.componentWillUnmount;

                  this.componentWillUnmount = function () {
                    willUnmountRef.apply(_this4);
                    clearTimeoutQueue(_this4.tckTimerQueue);
                  };
                }
              } else if (modifier === modifiers.ASYNC || modifier === modifiers.ASYNC + ".once") {
                tck = function tck() {
                  var _evts$eventId3;

                  isRC && (context = _objectSpread({}, _this4.state, _this4.props));
                  if (propKey && _this4["".concat(watchPropKey, "_").concat(ONCE)]) return;
                  if (stateKey && _this4["".concat(watchStateKey, "_").concat(ONCE)]) return;

                  (_evts$eventId3 = evts[eventId]).call.apply(_evts$eventId3, [null, context].concat(args));

                  once && propKey && (_this4["".concat(watchPropKey, "_").concat(ONCE)] = true);
                  once && stateKey && (_this4["".concat(watchStateKey, "_").concat(ONCE)] = true);
                };
              }

              propKey && (this.tckQueue[watchPropKey] = [tck]) && this.tckQueuePropKeys.push(watchPropKey);
              stateKey && (this.tckQueue[watchStateKey] = [tck]) && this.tckQueueStateKeys.push(watchStateKey); // 通过 react 生命周期函数 getSnapshotBeforeUpdate diff 值的变化控制埋点的上报

              if (isRC && !this.getSnapshotBeforeUpdate) {
                this.getSnapshotBeforeUpdate = function (prevProps, prevState) {
                  Object.keys(_this4.tckQueue).forEach(function (watchKey) {
                    var oldVal, newVal, cbks, key;

                    if (_this4.tckQueuePropKeys.includes(watchKey)) {
                      key = watchKey.split("_")[1];
                      newVal = _this4.props[key];
                      oldVal = prevProps[key];
                      cbks = _this4.tckQueue[watchKey];
                    }

                    if (_this4.tckQueueStateKeys.includes(watchKey)) {
                      key = watchKey.split("_")[1];
                      newVal = _this4.state[key];
                      oldVal = prevState[key];
                      cbks = _this4.tckQueue[watchKey];
                    }

                    oldVal !== newVal && cbks.forEach(function (sub) {
                      return sub && sub();
                    });
                  });
                  return null;
                };
              } else if (!isRC && this.reaction) {
                var cbks = this.tckQueue[watchStateKey];
                var disposer = this.reaction(function () {
                  return _this4[stateKey];
                }, function () {
                  cbks.forEach(function (sub) {
                    return sub && sub();
                  });
                  disposer();
                });
              }

              return fn.apply(void 0, args);
            };
          }

    if (value) {
      descriptor.value = function () {
        for (var _len6 = arguments.length, args = new Array(_len6), _key6 = 0; _key6 < _len6; _key6++) {
          args[_key6] = arguments[_key6];
        }

        return handler.apply(this, args);
      };
    } // 兼容箭头函数 https://github.com/MuYunyun/diana/issues/7


    if (initializer) {
      descriptor.initializer = function () {
        return handler.bind(this);
      };
    }
  };
}

export { inject, track };
